import json
from typing import Any

import requests


class Drive(object):
    def __init__(self, data=None, token: str = None, origin_uri: str = None):
        # TODO 2021-08-29 こっちでupload等の引数を受け取る様にする
        if data is None:
            data = {}
        else:
            data = json.loads(data)
            self.id = data.get('id')
            self.created_at = data.get('createdAt')
            self.name = data.get('name')
            self.type = data.get('type')
            self.md5 = data.get('md5')
            self.size = data.get('size')
            self.is_sensitive = data.get('isSensitive')
            self.blurhash = data.get('blurhash')
            self.properties = Properties(data.get('properties', {}))
            self.url = data.get('url')
            self.thumbnail_url = data.get('thumbnailUrl')
        self.token = token
        self.origin_uri = origin_uri
        self.path: str = ''
        self.name: str = ''

    def upload(self, path: str, name: str = None) -> 'Drive':
        """

        Parameters
        ----------
        path : str
            そのファイルまでのパスとそのファイル.拡張子(/home/test/test.png)
        name : str
            アップロードするファイルの名前
        Returns
        -------
        Drive: Drive
            upload後のレスポンスをDrive型に変更して返します
        """
        self.path = path
        self.name = name
        with open(path, 'rb') as f:
            file = f.read()
        args = {'i': f'{self.token}'}
        file = {'file': file}

        res = requests.post(self.origin_uri + '/api/drive/files/create', data=args, files=file)
        return Drive(res.text)


class Properties(object):
    def __init__(self, data=None):
        if data is None:
            data = {}
        self.width = data.get('width')
        self.height = data.get('height')



import re

from mi import Note, Drive


class API(object):
    def __init__(self, token, origin_uri):
        self.token = token
        if _origin_uri := re.search(r'wss?://(.*)/streaming', origin_uri):
            origin_uri = _origin_uri.group(0).replace('wss', 'https').replace('ws', 'http').replace('/streaming', '')

        if origin_uri[-1] == '/':
            self.origin_uri = origin_uri[:-1]
        else:
            self.origin_uri = origin_uri

    def note(self, *args, **kwargs):
        return Note(token=self.token, origin_uri=self.origin_uri, *args, **kwargs)

    def drive(self):
        return Drive(token=self.token, origin_uri=self.origin_uri)



import asyncio


class Loop(object):
    def __init__(self, func, seconds=None):
        self.seconds = seconds
        self.func = func
        self._task = None
        self.stop_next_iteration = None

    def start(self, *args, **kwargs) -> asyncio.Task:
        """
        タスクを開始する

        Parameters
        ----------
        args : Any
        kwargs : Any

        Returns
        -------
        _task : asyncio.Task
        """
        _loop = asyncio.get_event_loop()
        self._task = _loop.create_task(self.task(*args, **kwargs))
        return self._task

    def stop(self):
        """
        タスクを停止

        Returns
        -------
        None

        """
        if not self._task.done():
            self.stop_next_iteration = True

    async def task(self, *args, **kwargs):
        while True:
            if self.stop_next_iteration:
                return
            await self.func(self.seconds, *args, **kwargs)
            await asyncio.sleep(self.seconds)


def loop(n):
    def _deco(f):
        return Loop(f, n)

    return _deco



import json
import asyncio
from typing import Any

import websockets

from mi import Message, Reaction
from mi.router import Router


class WebSocket(object):
    """Misskey APIとやり取りを行うWebSocket object"""
    __slots__ = ['ws', 'cls', 'router']

    def __init__(self, cls):
        self.ws = None
        self.cls = cls
        self.router: Router

    async def _run(self, uri):
        try:
            async with websockets.connect(uri) as ws:
                asyncio.create_task(self._on_ready(ws))
                while True:
                    await ws.send(json.dumps({'type': 'connect',
                                              'body': {
                                                  'channel': 'globalTimeline',
                                                  'id': 'foobar',
                                                  'params': {
                                                      'some': 'thing'
                                                  }
                                              }}, ensure_ascii=False))
                    recv = await ws.recv()
                    asyncio.create_task(self._recv(ws, recv))
        except Exception as err:
            asyncio.create_task(self._on_error(err))

    async def _on_ready(self, ws):
        print('ready')
        self.router = Router(ws)
        await self.cls.on_ready(ws)

    async def _recv(self, ws: Any, message: Any):
        """

        Parameters
        ----------
        ws :
        message :
        """
        msg = Message(message, ws)
        event_list = {'note': '_on_message', 'reacted': '_on_reacted', 'deleted': '_on_deleted'}
        await getattr(self, f'{event_list.get(msg.header.type, "_on_message")}')(ws, message)

    async def _on_message(self, ws, message: Any) -> asyncio.Task:
        """
        Parameters
        ----------
        ws:
            WebSocket Instance
        message:
            Received message

        Returns
        -------
        task: asyncio.Task
        """
        message = Message(message, ws)
        await self.router.capture_message(message)
        if not hasattr(message.note, 'res'):
            task = asyncio.create_task(self.cls.on_message(ws, message))
        else:
            task = asyncio.create_task(self.cls.on_response(ws, message))

        return task

    async def _on_reacted(self, ws, message):
        asyncio.create_task(self.cls.on_reacted(ws, Reaction(message)))

    async def _on_deleted(self, ws, message):
        asyncio.create_task(self.cls.on_deleted(ws, Message(message, ws)))

    async def _on_error(self, err):
        await self.cls.on_error(err)

    async def _on_close(self, ws):
        pass



import json
import re
from typing import Any

import requests


class Message(object):
    def __init__(self, data: Any, ws=None):
        data = json.loads(data)
        self.type = data.get('type')
        self.header = Header(data.get('body', {}))
        if note := data.get('body', {}).get('body', None):
            self.note = Note(note, ws)
        elif note := data.get('createdNote', None):  # APIの場合
            self.note = Note(note, ws)
        else:
            data = data.get('body', {}).get('res', {}).get('createdNote', {})  # WebSocketsの場合
            data['res'] = True
            self.note = Note(data, ws)


class Header(object):
    def __init__(self, data):
        self.id = data.get('id')
        self.type = data.get('type')


class Note(object):
    def __init__(self, data=None, ws=None, text: str = '', cw: str = '', via_mobile: bool = False, token: str = None,
                 origin_uri: str = None):
        if data is None:  # リストをデフォルトにすると使いまわされて良くないので毎回初期化する必要がある。
            data = {}
        self.token = token
        self.ws = ws
        self.field = data
        self.origin_uri = origin_uri
        if not data:  # 型変更としてではなく、投稿などに使う際に必要
            self.field['text'] = text
            if len(cw) != 0:
                self.field['cw'] = cw
            self.field['viaMobile'] = via_mobile
            self.field['i'] = token
            self.field['mediaIds'] = []

        after_key = {'user': 'author'}
        for attr in ('id', 'createdAt', 'userId', 'user', 'text', 'cw',
                     'visibility', 'renoteCount', 'repliesCount', 'reactions',
                     'emojis', 'fileIds', 'files', 'replyId',
                     'renoteId', 'res'
                     ):
            try:
                value = data[attr]
                p = re.compile('[A-Z]')
                default_key = ("_" + p.search(attr)[0].lower()).join(p.split(attr)) if p.search(attr) is not None else attr
                key = after_key.get(default_key, default_key)
            except KeyError:
                continue
            else:  # エラーが発生しなかった場合は変数に追加
                if key == 'author':
                    setattr(self, key, User(value))
                else:
                    setattr(self, key, data[attr])

    def add_file(self, file_id: str = None, path: str = None, name: str = None) -> None:
        """
        ノートにファイルを添付します。

        Parameters
        ----------
        file_id : str
            既にドライブにあるファイルを使用する場合のファイルID
        path : str
            新しくファイルをアップロードする際のファイルへのパス
        name : str
            新しくファイルをアップロードする際のファイル名(misskey side

        Returns
        -------
        None
        """
        from mi import Drive
        if file_id is None:
            res = Drive(token=self.token, origin_uri=self.origin_uri).upload(path=path, name=name)
            print(res.id)
            self.field['mediaIds'].append(f'{res.id}')
        else:
            self.field['mediaIds'].append(f'{file_id}')

    async def send(self) -> requests.models.Response:
        print(self.field)
        data = json.dumps(self.field)
        res = requests.post(self.origin_uri + '/api/notes/create', data=data)
        print(res.text)
        msg = Message(res.text)
        return msg


class Reaction(object):
    def __init__(self, data):
        data = json.loads(data)
        self.header = Header(data.get('body', {}))
        self.note = ReactionNote(data.get('body'))


class ReactionNote(object):
    def __init__(self, data):
        self.reaction = data['body'].get('reaction')
        self.user_id = data['body'].get('userId')


class User(object):
    __slots__ = (
        'id',
        'name',
        'username',
        'host',
        'avatar_url',
        'avatar_blurhash',
        'avatar_color',
        'instance',
        'emojis'
    )

    def __init__(self, data: dict):
        self.id = data.get('id')
        self.name = data.get('name', data.get('username', None))
        self.username = data.get('username')
        self.host = data.get('host')
        self.avatar_url = data.get('avatarUrl')
        self.avatar_blurhash = data.get('avatarBlurhash')
        self.avatar_color = data.get('avatarColor')
        self.instance = Instance(data)
        self.emojis = data.get('emojis')


class Instance(object):
    __slots__ = (
        'home',
        'name',
        'software_name',
        'icon_url',
        'favicon_url',
        'theme_color'
    )

    def __init__(self, data):
        self.home = data.get('instance', {}).get('home')
        self.name = data.get('instance', {}).get('name')
        self.software_name = data.get('instance', {}).get('softwareName')
        self.icon_url = data.get('instance', {}).get('iconUrl')
        self.favicon_url = data.get('instance', {}).get('faviconUrl')
        self.theme_color = data.get('instance', {}).get('themeColor')



import json

from mi import Message


class Router(object):
    def __init__(self, ws):
        self.ws = ws

    async def channels(self, channel_list: list):
        channel_dict = {'global': self.global_time_line, 'main': self.main_channel,
                        'home': self.home_time_line, 'local': self.local_time_line}
        try:
            for channel in channel_list:
                func = channel_dict.get(channel)
                await getattr(self, func.__name__)()
        except KeyError:
            pass

    async def global_time_line(self):
        await self.ws.send(json.dumps({
            'type': 'connect',
            'body': {
                'channel': 'globalTimeline',
                'id': 'foobar',
                'params': {
                    'some': 'thing'
                }
            }
        }, ensure_ascii=False))

    async def main_channel(self):
        await self.ws.send(json.dumps({
            'type': 'connect',
            'body': {
                'channel': 'main',
                'id': 'foobar',
                'params': {
                    'some': 'thing'
                }
            }
        }, ensure_ascii=False))

    async def home_time_line(self):
        await self.ws.send(json.dumps({
            'type': 'connect',
            'body': {
                'channel': 'homeTimeline',
                'id': 'foobar',
                'params': {
                    'some': 'thing'
                }
            }
        }, ensure_ascii=False))

    async def local_time_line(self):
        await self.ws.send(json.dumps({
            'type': 'connect',
            'body': {
                'channel': 'localTimeline',
                'id': 'oobar',
                'params': {
                    'some': 'ting'
                }
            }
        }, ensure_ascii=False))

    async def capture_message(self, message: Message):
        if hasattr(message, 'id'):
            await self.ws.send(json.dumps({
                'type': 'subNote',
                'body': {
                    'id': f'{message.note.id}'
                }
            }))



import asyncio
import importlib
import sys
import traceback
from typing import Any, Callable, Coroutine

from mi.http import WebSocket


class BotBase(WebSocket):
    def __init__(self):
        self.extra_events = {}
        self.special_events = {}
        self.token = None
        super().__init__(self)

    def event(self, name=None):
        def decorator(func):
            self.add_event(func, name)
            return func

        return decorator

    def add_event(self, func, name=None):
        name = func.__name__ if name is None else name
        if not asyncio.iscoroutinefunction(func):
            raise TypeError('Listeners must be coroutines')

        if name in self.extra_events:
            self.special_events[name].append(func)
        else:
            self.special_events[name] = [func]

    def listen(self, name=None):

        def decorator(func):
            self.add_listener(func, name)
            return func

        return decorator

    def add_listener(self, func, name=None):
        name = func.__name__ if name is None else name
        if not asyncio.iscoroutinefunction(func):
            raise TypeError('Listeners must be coroutines')

        if name in self.extra_events:
            self.extra_events[name].append(func)
        else:
            self.extra_events[name] = [func]

    async def event_dispatch(self, event_name, *args, **kwargs):
        ev = 'on_' + event_name
        for event in self.special_events.get(ev, []):
            foo = importlib.import_module(event.__module__)
            coro = getattr(foo, ev)
            await self.schedule_event(coro, event, *args, **kwargs)

    async def dispatch(self, event_name, *args, **kwargs):
        ev = 'on_' + event_name
        for event in self.extra_events.get(ev, []):
            foo = importlib.import_module(event.__module__)
            coro = getattr(foo, ev)
            await self.schedule_event(coro, event, *args, **kwargs)

    async def schedule_event(self, coro: Callable[..., Coroutine[Any, Any, Any]], event_name: str, *args: Any,
                             **kwargs: Any) -> asyncio.Task:
        return asyncio.create_task(self._run_event(coro, event_name, *args, **kwargs), name=f'MI.py: {event_name}')

    async def _run_event(self, coro: Callable[..., Coroutine[Any, Any, Any]], event_name: str, *args: Any,
                         **kwargs: Any) -> None:
        try:
            await coro(*args, **kwargs)
        except asyncio.CancelledError:
            pass
        except Exception:
            try:
                await self.__on_error(event_name, *args, **kwargs)
            except asyncio.CancelledError:
                pass

    async def __on_error(self, event_method: str) -> None:
        print(f'Ignoring exception in {event_method}', file=sys.stderr)
        traceback.print_exc()

    async def on_error(self, err):
        await self.event_dispatch('error', err)

    async def on_message(self, ws, message):
        """デフォルト処理"""
        await self.dispatch('message', ws, message)

    async def on_response(self, ws, message):
        await self.dispatch('response', ws, message)

    async def on_ready(self, ws):
        await self.event_dispatch('ready', ws)

    async def on_reacted(self, ws, message):
        await self.dispatch('reaction', ws, message)

    async def on_deleted(self, ws, message):
        await self.dispatch('deleted', ws, message)

    def run(self, uri: str, token: str) -> None:
        """
        Launch the bot.
        Parameters
        ----------
        uri : str
            websocket url of the Misskey instance to connect to
        token : str
            Misskey account token

        None

        Examples
        --------

        When inheriting from a class: ::

            class MyBot(Bot):
                async def on_message(self, ws, message):
                    pass
            bot = MyBot()
            bot.run(uri, token)

        When using a listener: ::

            bot = Bot()

            @bot.event()
            async def on_message(ws, message):
                pass

            bot.run(uri, token)
        """
        self.token = token
        asyncio.get_event_loop().run_until_complete(self._run(f'{uri}?i={token}'))


class Bot(BotBase):
    pass



from setuptools import find_packages, setup
import pathlib


description = 'A Python wrapper for the Misskey API'
readme_file = pathlib.Path(__file__).parent/'README.md'
with readme_file.open(encoding='utf-8') as fh:
    long_description = fh.read()

setup(
    name='mi.py',
    version='0.1.0',
    install_requires=['websockets', 'requests'],
    url='https://github.com/yupix/mi.py',
    author='yupix',
    author_email='yupi0982@outlook.jp',
    license='MIT',
    python_requires='>=3.8, <4.0',
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Programming Language :: Python :: 3.9',
        'Natural Language :: Japanese',
        'License :: OSI Approved :: MIT License',
    ]
)



__title__ = 'misskey'
__author__ = 'yupix'
__license__ = 'MIT'
__copyright__ = 'Copyright 2021-present yupix'
__author_email__ = 'yupi0982@outlook.jp'
__url__ = 'https://github.com/yupix/Mi.py'
__version__ = '0.1.0'

__path__ = __import__('pkgutil').extend_path(__path__, __name__)

from .ext import task
from .ext.task import Loop, loop
from .note import Message, Header, Note, Reaction, ReactionNote, User, Instance
from .router import Router
from .utils import bool_to_string
from .drive import Properties, Drive
from .bot import Bot, BotBase

__all__ = [
    'bot',
    'http',
    'note',
    'router',
    'ext',
    'task',
    'Loop',
    'loop',
    'utils',
    'drive',
    'Note',
    'ReactionNote',
    'Instance',
    'bool_to_string',
    'Message',
    'Header',
    'User',
    'Reaction',
    'Bot',
    'Router',
    'Drive',
    'Properties'
]



def bool_to_string(boolean: bool) -> str:
    """
    boolを小文字にして文字列として返します

    Parameters
    ----------
    boolean : bool
        変更したいbool値
    Returns
    -------
    true or false: str
        小文字になったbool文字列
    """

    return 'true' if boolean else 'false'
